import * as crypto from "crypto"
import { IFileSystem, OnePasswordFileManager } from "./fs"

import type { Profile, Band, Overview, EncryptedItem, Item } from "./types"

type FoldersMap = { [uuid: string]: Band }

/**
 * Main OnePassword Vault class
 */
export class Vault {
  // Ciphers
  #master: Cipher
  #overview: Cipher

  // File system interface
  #files: OnePasswordFileManager

  // Encrypted contents
  #profile: Profile
  #folders: FoldersMap
  #bands = new Map<string, Band>()

  // Decrypted contents with plain-texts
  #overviews = new Map<string, Overview>()
  #items = new Map<string, EncryptedItem>()

  private constructor(
    files: OnePasswordFileManager,
    profile: Profile,
    folders: FoldersMap,
    bands: Map<string, Band>
  ) {
    this.#files = files
    this.#profile = profile
    this.#folders = folders
    this.#bands = bands

    this.#files
    this.#folders
    this.#items
  }

  /**
   * Create a new OnePassword Vault instance and read all bands.
   */
  static async of(path: string, profileName = "default", fs: IFileSystem) {
    const files = new OnePasswordFileManager(fs, path, profileName)
    const profile = JSON.parse(stripText(await files.getProfile(), "var profile=", ";"))
    const folders = JSON.parse(stripText(await files.getFolders(), "loadFolders(", ");"))
    const bands = new Map<string, Band>()

    for (let i = 0; i < 36; i++) {
      const letter = i.toString(36)
      const source = await files.getBand(letter)
      bands.set(letter, source ? JSON.parse(stripText(source, "ld(", ");")) : {})
    }

    return new Vault(files, profile, folders, bands)
  }

  /**
   * Unlock this OnePassword vault.
   * @param masterPassword User provided master password. Only the derived
   * master and overview key will be stored within the class.
   */
  unlock(masterPassword: string) {
    const profile = this.#profile
    const derivedKey = crypto.pbkdf2Sync(
      masterPassword,
      toBuffer(profile.salt),
      profile.iterations,
      /* keylen */ 64,
      "sha512"
    )

    const cipher = splitPlainText(derivedKey)

    // Derive master key and overview keys
    try {
      this.#master = decryptKeys(profile.masterKey, cipher)
      this.#overview = decryptKeys(profile.overviewKey, cipher)
    } catch (e) {
      if (e instanceof HMACAssertionError) {
        throw new Error("Invalid password.")
      }
      throw e
    }
  }

  /**
   * Remove derived keys within the class instance.
   */
  lock() {
    this.#master = null!
    this.#overview = null!
    this.#overviews.clear()
  }

  get isLocked() {
    return Boolean(this.#master.key)
  }

  getOverview(title: string): Overview
  getOverview(predicate: (overview: Overview) => boolean): Overview

  getOverview(condition: string | ((overview: Overview) => boolean)) {
    if (typeof condition === "string") {
      const title = condition
      condition = overview => overview.title === title
    }
    for (const value of this.#overviews.values()) {
      if (condition(value)) {
        return value
      }
    }
  }

  getItem(uuid: string) {
    const encrypted = uuid ? this.#bands.get(uuid[0])![uuid] : undefined
    return encrypted && this.decryptItem(encrypted)
  }

  readItems() {
    this.#bands.forEach(value => {
      for (const [uuid, item] of Object.entries(value)) {
        const overview = this.decryptOverview(item)
        overview.uuid = uuid
        this.#overviews.set(uuid, overview)
      }
    })
    return this
  }

  private decryptOverview(item: EncryptedItem) {
    try {
      const overview = decryptOPData(toBuffer(item.o), this.#overview)
      return JSON.parse(overview.toString("utf8")) as Overview
    } catch (e) {
      console.error("Failed to decrypt overview item.")
      throw e
    }
  }

  private decryptItem(item: EncryptedItem): Item {
    const k = toBuffer(item.k)
    const data = k.slice(0, -32)
    assertHMac(data, this.#master.hmac, k.slice(-32))
    const derivedKey = decryptData(this.#master.key, data.slice(0, 16), data.slice(16))

    const detail = decryptOPData(
      /* cipherText */ toBuffer(item.d),
      /* cipher */ splitPlainText(derivedKey)
    )
    return JSON.parse(detail.toString("utf-8"))
  }
}

/** Encryption and MAC */
interface Cipher {
  /** Encryption key */
  key: Buffer
  /** HMAC key */
  hmac: Buffer
}

function stripText(text: string, prefix: string, suffix: string) {
  if (text.startsWith(prefix)) text = text.slice(prefix.length)
  if (text.endsWith(suffix)) text = text.slice(0, -suffix.length)
  return text
}

const splitPlainText = (derivedKey: Buffer): Cipher => ({
  key: derivedKey.slice(0, 32),
  hmac: derivedKey.slice(32, 64),
})

class HMACAssertionError extends Error {}

function decryptOPData(cipherText: Buffer, cipher: Cipher) {
  const key = cipherText.slice(0, -32)
  assertHMac(key, cipher.hmac, cipherText.slice(-32))

  const plaintext = decryptData(cipher.key, key.slice(16, 32), key.slice(32))
  const size = readUint16(key.slice(8, 16))
  return plaintext.slice(-size)
}

function assertHMac(data: Buffer, key: Buffer, expected: Buffer) {
  const actual = crypto.createHmac("sha256", key).update(data).digest()
  if (!actual.equals(expected)) {
    throw new HMACAssertionError("HMAC assertion failed.")
  }
}

function decryptKeys(encryptedKey: string, derived: Cipher) {
  const buffer = toBuffer(encryptedKey)
  const base = decryptOPData(buffer, derived)
  const digest = crypto.createHash("sha512").update(base).digest()
  return splitPlainText(digest)
}

function decryptData(key: Buffer, iv: Buffer, data: Buffer) {
  const cipher = crypto.createDecipheriv("aes-256-cbc", key, iv).setAutoPadding(false)
  return Buffer.concat([cipher.update(data), cipher.final()])
}

function readUint16({ buffer, byteOffset, length }: Buffer) {
  return new DataView(buffer, byteOffset, length).getUint16(0, true)
}

function toBuffer(data: string) {
  return Buffer.from(data, "base64")
}
